# type : ignore
from dataclasses import dataclass
from typing import List, Union

import torch
import torch.nn.functional as F
from kodali import NerTags
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast

from luie.labels.named_entity_recognition.korean_corpus import NerKoreanCorpusLabels
from luie.models.sequence_labeling import SequenceLabelingModel
from luie.tokenizers import KoCharElectraTokenizer

TOKENIZER_NAME = "monologg/kocharelectra-base-discriminator"

HF_MODEL_NAME = "chnaaam/kocharelectra-ner"


@dataclass
class NerOutput:
    word: str
    label: str
    start_idx: int
    end_idx: int


def decode(
    tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
    token_ids: List[int],
    labels: List[str],
) -> List[NerOutput]:
    outputs = []
    buffer = {}

    for idx, label in enumerate(labels):
        if label.startswith(str(NerTags.SINGLE)):
            outputs.append(
                NerOutput(
                    word=tokenizer.decode(token_ids=token_ids[idx]),
                    label=label.replace(f"{str(NerTags.SINGLE)}", ""),
                    start_idx=idx,
                    end_idx=idx,
                )
            )
        if label.startswith(str(NerTags.BEGIN)):
            buffer = {
                "idx": idx,
                "label": label.replace(f"{str(NerTags.BEGIN)}-", ""),
            }
        elif label.startswith(str(NerTags.END)):
            start_idx = buffer["idx"]
            label = buffer["label"]

            outputs.append(
                NerOutput(
                    word=tokenizer.decode(token_ids=token_ids[start_idx : idx + 1]),
                    label=label,
                    start_idx=start_idx,
                    end_idx=idx,
                )
            )

            # Reset
            start_idx = -1
            label = str(NerTags.OUTSIDE)

    return outputs


class NerInferenceEngineModule:
    def __init__(self, device_id: int = 0) -> None:
        self.device = f"cuda:{device_id}" if device_id != -1 else "cpu"

        if not torch.cuda.is_available():
            self.device = "cpu"

        self.model = SequenceLabelingModel(model_name=HF_MODEL_NAME)
        self.model.eval()
        self.model = self.model.to(self.device)  # type: ignore

        self.tokenizer = KoCharElectraTokenizer.from_pretrained(TOKENIZER_NAME)

    def run(self, sentence: str) -> List[NerOutput]:
        encoded_inputs = self.tokenizer(sentence, return_tensors="pt")
        inputs = {k: v.to(self.device) for k, v in encoded_inputs.items()}

        outputs = self.model(**inputs)
        outputs = torch.argmax(F.softmax(outputs.logits, dim=-1), dim=-1).tolist()[0]

        outputs = decode(
            tokenizer=self.tokenizer,
            token_ids=encoded_inputs["input_ids"].tolist()[0][1:-1],  # Skip CLS and SEP tokens,
            labels=[NerKoreanCorpusLabels.IDX2LABEL[str(o)] for o in outputs][1:-1],  # Skip CLS and SEP tokens,
        )

        return outputs



# AI Hub
from typing import Dict


class ReAiHubLabels:
    SIZE: int = 20
    LABEL2IDX: Dict[str, int] = {
        "isParentOf": 0,
        "isUsedFor": 1,
        "isOpponentOf": 2,
        "no-Relation": 3,
        "isSiblingOf": 4,
        "isPlaceOf": 5,
        "isChildOf": 6,
        "isMemberOf": 7,
        "isSpouseOf": 8,
        "isToBe": 9,
        "hasAttribute": 10,
        "inFavorOf ": 11,
        "isAgainst": 12,
        "isA": 13,
        "isColleagueOf": 14,
        "isPartOf": 15,
        "isCausedBy": 16,
        "isTimeOf": 17,
        "isRelativeOf": 18,
        "inFavorOf": 19,
    }
    IDX2LABEL: Dict[str, str] = {
        "0": "isParentOf",
        "1": "isUsedFor",
        "2": "isOpponentOf",
        "3": "no-Relation",
        "4": "isSiblingOf",
        "5": "isPlaceOf",
        "6": "isChildOf",
        "7": "isMemberOf",
        "8": "isSpouseOf",
        "9": "isToBe",
        "10": "hasAttribute",
        "11": "inFavorOf ",
        "12": "isAgainst",
        "13": "isA",
        "14": "isColleagueOf",
        "15": "isPartOf",
        "16": "isCausedBy",
        "17": "isTimeOf",
        "18": "isRelativeOf",
        "19": "inFavorOf",
    }



from transformers import AutoModelForSequenceClassification, PreTrainedModel


class TextClassificationModel:
    def __new__(cls, model_name: str, num_labels: int = -1) -> PreTrainedModel:
        params = {"pretrained_model_name_or_path": model_name, "num_labels": num_labels}

        if num_labels == -1:
            del params["num_labels"]

        return AutoModelForSequenceClassification.from_pretrained(**params)



# type: ignore
# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team and Jangwon Park
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tokenization classes."""

import collections
import logging
import os
import unicodedata
from typing import List, Optional

from transformers.tokenization_utils import PreTrainedTokenizer

logger = logging.getLogger(__name__)

VOCAB_FILES_NAMES = {"vocab_file": "vocab.txt"}

PRETRAINED_VOCAB_FILES_MAP = {}

PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES = {}

PRETRAINED_INIT_CONFIGURATION = {}


def load_vocab(vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = collections.OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n")
        vocab[token] = index
    return vocab


def whitespace_tokenize(text):
    """Runs basic whitespace cleaning and splitting on a piece of text."""
    text = text.strip()
    if not text:
        return []
    tokens = text.split()
    return tokens


class KoCharElectraTokenizer(PreTrainedTokenizer):
    r"""
    Korean Character Electra tokenizer. Character-level Tokenizer
    This tokenizer inherits from :class:`~transformers.PreTrainedTokenizer` which contains most of the methods. Users
    should refer to the superclass for more information regarding methods.
    Args:
        vocab_file (:obj:`string`):
            File containing the vocabulary.
        do_lower_case (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Whether to lowercase the input when tokenizing.
        do_basic_tokenize (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Whether to do basic tokenization before WordPiece.
        never_split (:obj:`bool`, `optional`, defaults to :obj:`True`):
            List of tokens which will never be split during tokenization. Only has an effect when
            :obj:`do_basic_tokenize=True`
        unk_token (:obj:`string`, `optional`, defaults to "[UNK]"):
            The unknown token. A token that is not in the vocabulary cannot be converted to an ID and is set to be this
            token instead.
        sep_token (:obj:`string`, `optional`, defaults to "[SEP]"):
            The separator token, which is used when building a sequence from multiple sequences, e.g. two sequences
            for sequence classification or for a text and a question for question answering.
            It is also used as the last token of a sequence built with special tokens.
        pad_token (:obj:`string`, `optional`, defaults to "[PAD]"):
            The token used for padding, for example when batching sequences of different lengths.
        cls_token (:obj:`string`, `optional`, defaults to "[CLS]"):
            The classifier token which is used when doing sequence classification (classification of the whole
            sequence instead of per-token classification). It is the first token of the sequence when built with
            special tokens.
        mask_token (:obj:`string`, `optional`, defaults to "[MASK]"):
            The token used for masking values. This is the token used when training this model with masked language
            modeling. This is the token which the model will try to predict.
        tokenize_chinese_chars (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Whether to tokenize Chinese characters.
            This should likely be deactivated for Japanese:
            see: https://github.com/huggingface/transformers/issues/328
    """
    vocab_files_names = VOCAB_FILES_NAMES
    pretrained_vocab_files_map = PRETRAINED_VOCAB_FILES_MAP
    pretrained_init_configuration = PRETRAINED_INIT_CONFIGURATION
    max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES

    def __init__(
        self,
        vocab_file,
        do_lower_case=False,
        do_basic_tokenize=True,
        never_split=None,
        unk_token="[UNK]",
        sep_token="[SEP]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        mask_token="[MASK]",
        tokenize_chinese_chars=True,
        **kwargs
    ):
        super().__init__(
            unk_token=unk_token,
            sep_token=sep_token,
            pad_token=pad_token,
            cls_token=cls_token,
            mask_token=mask_token,
            **kwargs,
        )

        if not os.path.isfile(vocab_file):
            raise ValueError(
                "Can't find a vocabulary file at path '{}'. To load the vocabulary from a Google pretrained "
                "model use `tokenizer = BertTokenizer.from_pretrained(PRETRAINED_MODEL_NAME)`".format(vocab_file)
            )
        self.vocab = load_vocab(vocab_file)
        self.ids_to_tokens = collections.OrderedDict([(ids, tok) for tok, ids in self.vocab.items()])

    @property
    def vocab_size(self):
        return len(self.vocab)

    def get_vocab(self):
        return dict(self.vocab, **self.added_tokens_encoder)

    def _tokenize(self, text):
        return list(" ".join(text.split()))  # Erase duplicate space

    def _convert_token_to_id(self, token):
        """Converts a token (str) in an id using the vocab."""
        return self.vocab.get(token, self.vocab.get(self.unk_token))

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (str) using the vocab."""
        return self.ids_to_tokens.get(index, self.unk_token)

    def convert_tokens_to_string(self, tokens):
        """Converts a sequence of tokens (string) in a single string."""
        return "".join(tokens).strip()

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks
        by concatenating and adding special tokens.
        A BERT sequence has the following format:
        - single sequence: ``[CLS] X [SEP]``
        - pair of sequences: ``[CLS] A [SEP] B [SEP]``
        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs to which the special tokens will be added
            token_ids_1 (:obj:`List[int]`, `optional`, defaults to :obj:`None`):
                Optional second list of IDs for sequence pairs.
        Returns:
            :obj:`List[int]`: list of `input IDs <../glossary.html#input-ids>`__ with the appropriate special tokens.
        """
        if token_ids_1 is None:
            return [self.cls_token_id] + token_ids_0 + [self.sep_token_id]
        cls = [self.cls_token_id]
        sep = [self.sep_token_id]
        return cls + token_ids_0 + sep + token_ids_1 + sep

    def get_special_tokens_mask(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None, already_has_special_tokens: bool = False
    ) -> List[int]:
        """
        Retrieves sequence ids from a token list that has no special tokens added. This method is called when adding
        special tokens using the tokenizer ``prepare_for_model`` or ``encode_plus`` methods.
        Args:
            token_ids_0 (:obj:`List[int]`):
                List of ids.
            token_ids_1 (:obj:`List[int]`, `optional`, defaults to :obj:`None`):
                Optional second list of IDs for sequence pairs.
            already_has_special_tokens (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Set to True if the token list is already formatted with special tokens for the model
        Returns:
            :obj:`List[int]`: A list of integers in the range [0, 1]: 0 for a special token, 1 for a sequence token.
        """

        if already_has_special_tokens:
            if token_ids_1 is not None:
                raise ValueError(
                    "You should not supply a second sequence if the provided sequence of "
                    "ids is already formated with special tokens for the model."
                )
            return list(map(lambda x: 1 if x in [self.sep_token_id, self.cls_token_id] else 0, token_ids_0))

        if token_ids_1 is not None:
            return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1]

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Creates a mask from the two sequences passed to be used in a sequence-pair classification task.
        A BERT sequence pair mask has the following format:
        ::
            0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1
            | first sequence    | second sequence |
        if token_ids_1 is None, only returns the first portion of the mask (0's).
        Args:
            token_ids_0 (:obj:`List[int]`):
                List of ids.
            token_ids_1 (:obj:`List[int]`, `optional`, defaults to :obj:`None`):
                Optional second list of IDs for sequence pairs.
        Returns:
            :obj:`List[int]`: List of `token type IDs <../glossary.html#token-type-ids>`_ according to the given
            sequence(s).
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]
        return len(cls + token_ids_0 + sep) * [0] + len(token_ids_1 + sep) * [1]

    def save_vocabulary(self, vocab_path):
        """
        Save the sentencepiece vocabulary (copy original file) and special tokens file to a directory.
        Args:
            vocab_path (:obj:`str`):
                The directory in which to save the vocabulary.
        Returns:
            :obj:`Tuple(str)`: Paths to the files saved.
        """
        index = 0
        if os.path.isdir(vocab_path):
            vocab_file = os.path.join(vocab_path, VOCAB_FILES_NAMES["vocab_file"])
        else:
            vocab_file = vocab_path
        with open(vocab_file, "w", encoding="utf-8") as writer:
            for token, token_index in sorted(self.vocab.items(), key=lambda kv: kv[1]):
                if index != token_index:
                    logger.warning(
                        "Saving vocabulary to {}: vocabulary indices are not consecutive."
                        " Please check that the vocabulary is not corrupted!".format(vocab_file)
                    )
                    index = token_index
                writer.write(token + "\n")
                index += 1
        return (vocab_file,)



# KLUE
from typing import Dict


class ReKlueLabels:
    SIZE: int = 30
    LABEL2IDX: Dict[str, int] = {
        "no_relation": 0,
        "org:dissolved": 1,
        "org:founded": 2,
        "org:place_of_headquarters": 3,
        "org:alternate_names": 4,
        "org:member_of": 5,
        "org:members": 6,
        "org:political/religious_affiliation": 7,
        "org:product": 8,
        "org:founded_by": 9,
        "org:top_members/employees": 10,
        "org:number_of_employees/members": 11,
        "per:date_of_birth": 12,
        "per:date_of_death": 13,
        "per:place_of_birth": 14,
        "per:place_of_death": 15,
        "per:place_of_residence": 16,
        "per:origin": 17,
        "per:employee_of": 18,
        "per:schools_attended": 19,
        "per:alternate_names": 20,
        "per:parents": 21,
        "per:children": 22,
        "per:siblings": 23,
        "per:spouse": 24,
        "per:other_family": 25,
        "per:colleagues": 26,
        "per:product": 27,
        "per:religion": 28,
        "per:title": 29,
    }
    IDX2LABEL: Dict[str, str] = {
        "0": "no_relation",
        "1": "org:dissolved",
        "2": "org:founded",
        "3": "org:place_of_headquarters",
        "4": "org:alternate_names",
        "5": "org:member_of",
        "6": "org:members",
        "7": "org:political/religious_affiliation",
        "8": "org:product",
        "9": "org:founded_by",
        "10": "org:top_members/employees",
        "11": "org:number_of_employees/members",
        "12": "per:date_of_birth",
        "13": "per:date_of_death",
        "14": "per:place_of_birth",
        "15": "per:place_of_death",
        "16": "per:place_of_residence",
        "17": "per:origin",
        "18": "per:employee_of",
        "19": "per:schools_attended",
        "20": "per:alternate_names",
        "21": "per:parents",
        "22": "per:children",
        "23": "per:siblings",
        "24": "per:spouse",
        "25": "per:other_family",
        "26": "per:colleagues",
        "27": "per:product",
        "28": "per:religion",
        "29": "per:title",
    }



from typing import List

from luie.modules.ner_module import NerInferenceEngineModule, NerOutput


class LuieEngine:
    def __init__(self, task: str, device_id: int = 0, use_onnx: bool = False) -> None:
        if task == "ner":
            self.module = NerInferenceEngineModule()
        else:
            raise ValueError("Unsupported task. 'ner' task is currently supported.")

    def run(self, sentence: str) -> List[NerOutput]:
        return self.module.run(sentence=sentence)
